#             Autore: Roberto FULIGNI
#
#               Data: 28/11/2023
#    
#        Descrizione: Esempio di codifica di una stringa mediante algoritmo di Huffman

def codifica_frase(f):
    # Step 1: calcola le frequenze dei caratteri presenti in f
    freq = {}
    for c in f:
        freq[c] = freq.get(c, 0) + 1

    tabella = {}

    # Caso particolare: se la frase contiene un solo carattere (eventualmente ripetuto),
    #                   la tabella è costitutita da un unico carattere di codice "0" 
    if (len(freq) == 1):
        for c in freq:
            tabella[c] = "0"
    else:
        # Inizialmente tutti i caratteri sono privi di codifica
        for c in freq:
            tabella[c] = ""

    # Step 2: determina una lista provvisoria di elementi. Ogni elemento è composto
    #         da una coppia di dati: il carattere e la sua frequenza

    lista = []
    for c in freq:
        lista.append((c, freq[c]))

    # Step 3: si applica ripetutamente l'algoritmo di Huffman per ridurre 
    #         progressivamente la dimensione della lista. L'algoritmo termina
    #         quando la lista contiene una sola coppia

    while len(lista) > 1:
        # Step 3a: si ordina la lista per valori decrescenti di frequenza,
        #          a parità di frequenza si ordina per carattere
        lista.sort(key = lambda x: (-x[1], x[0])) 
        
        # Step 3b: Si estraggono (pop) dalla lista ordinata le ultime due coppie 
        #          (quelle con frequenze minori) e si aggionaro le codifiche di 
        #          tutti i caratteri presenti nelle coppie
        #  
        coppia1 = lista.pop(-1)     # -1 -> ultima posizione
        for c1 in coppia1[0]:
            # Aggiungiamo un 1 in testa a tutti caratteri dell'ultima coppia...
            tabella[c1] = "1" + tabella[c1]

        coppia2 = lista.pop(-1)
        for c2 in coppia2[0]:
            # ... E uno 0 ai caratteri della penultima coppia
            tabella[c2] = "0" + tabella[c2]

        # Step 3c: si uniscono le due coppie concatenando i loro caratteri e 
        #          sommando le loro frequenze, il risultato è una nuova
        #          coppia che viene reinserita in coda alla lista

        coppia3 = (coppia1[0]+coppia2[0], coppia1[1]+coppia2[1])
        lista.append(coppia3)

    # Step 4: si utilizza la tabella costruita per codificare la frase in input

    risultato = ""
    for c in f:
        risultato = risultato + tabella[c]

    return (risultato, tabella, freq)

# Stampa la tabella delle frequenze e delle codifiche dei singoli caratteri
def stampa_tabella(t, f):
    print("\nCARATTERE  FREQUENZA  CODIFICA")
    print("------------------------------")
    for k in sorted(f.keys(), key=lambda x: (-f[x], x)):
        print(f"    {k}         {f[k]:3}       {t[k]}")

# Decodifica una sequenza binaria usando una data tabella di caratteri e relative codifiche
def decodifica_frase(cod, t):
    # Step 1: si inizializza il risultato, si effettua la decodifica di una carattere alla volta
    risultato = ""
    
    while len(cod) > 0:
        trovato = False
        # Cerca nella tabella il carattere la cui codifica coincide con la parte iniziale
        # della sequenza binaria
        for c in t:
            if cod.startswith(t[c]):
                # Il carattere è stato individuato, lo si aggiunge in coda al risultato
                risultato = risultato + c
                n = len(t[c])   # n = Lunghezza del codice trovato
                cod = cod[n:]   # Si rimuovono i primi n caratteri della sequenza
                trovato = True
        
        if not trovato:
            print("Impossibile decodificare la sequenza binaria")
            return None
    
    return risultato


################################################################
### PROGRAMMA PRINCIPALE: CODIFICA E DECODIFICA DI UNA FRASE ###
################################################################

frase = input("Frase da codificare: ")

(risultato, tabella, frequenze) = codifica_frase(frase)
stampa_tabella(tabella, frequenze)
print("\nRisultato della codifica di Huffman:", risultato)
print("Lunghezza del risultato:", len(risultato), "bit")

codice = risultato
print("\nUsiamo la stessa tabella per decodificare", codice)
risultato = decodifica_frase(codice, tabella)
print("La frase decodificata è:", risultato)
